/*
 *  Copyright (c) 1999 Sascha Schumann. All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 * 
 *  3. All advertising materials mentioning features or use of this
 *     software must display the following acknowledgment:
 *     "This product includes software developed by
 *      Sascha Schumann <ss@2ns.de>."
 * 
 *  4. Redistributions of any form whatsoever must retain the following
 *     acknowledgment:
 *     "This product includes software developed by
 *      Sascha Schumann <ss@2ns.de>."
 * 
 *  THIS SOFTWARE IS PROVIDED BY SASCHA SCHUMANN ``AS IS'' AND ANY
 *  EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL SASCHA SCHUMANN OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 *  OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
   $Id: mhash.c,v 1.10 1999/05/21 17:49:33 sascha Exp $ 
 */

#include <stdlib.h>

#include "libdefs.h"
#include "mhash.h"
#include "mhash_crc32.h"
#include "mhash_haval.h"
#include "mhash_md5.h"
#include "mhash_sha1.h"
#include "mhash_tiger.h"
#include "rmd128.h"
#include "rmd160.h"
#include "gosthash.h"

#define MAX_THREADS 256			/*
								   After that pointers are being reused 
								 */

/*
   I think now it is re-entrant 
 */
static int check_pointer[MAX_THREADS];
static int used_all_threads;
static int cur_thread;

#ifdef HAVE_LIBPTHREAD

#include <pthread.h>

pthread_mutex_t counter_cur_thread = PTHREAD_MUTEX_INITIALIZER;

#define CURRENT_COUNTER_LOCK pthread_mutex_lock(&counter_cur_thread)
#define CURRENT_COUNTER_UNLOCK pthread_mutex_unlock(&counter_cur_thread)


#else

#define CURRENT_COUNTER_LOCK
#define CURRENT_COUNTER_UNLOCK

#endif

static MD5_CTX *md5;			/*
								   For md5 
								 */
static SHA1_CTX *sha1;			/*
								   For sha1 
								 */
static word32 *crc;				/*
								   For crc32 
								 */
static haval_state *haval;		/*
								   For HAVAL 
								 */
static word32 *rmd128;
static word32 *rmd160;
static word64 *tiger_data;
static GostHashCtx *gost;

/*
   hold the key and the algorithm 
 */
static hashid *algorithm_given;

#define MHASH_ENTRY(name, blksize) { #name, name, blksize }

static mhash_hash_entry algorithms[] =
{
	MHASH_ENTRY(CRC32, 4),
	MHASH_ENTRY(MD5, 16),
	MHASH_ENTRY(SHA1, 20),
	MHASH_ENTRY(HAVAL, HAVAL_FPTLEN >> 3),
	MHASH_ENTRY(RIPEMD160, 160 >> 3),
	MHASH_ENTRY(RIPEMD128, 128 >> 3),
	MHASH_ENTRY(TIGER, 192 >> 3),
	MHASH_ENTRY(GOST, 32),
	MHASH_ENTRY(CRC32B, 4),
	{0}
};

#define MHASH_LOOP(b) \
	mhash_hash_entry *p; \
	for(p = algorithms; p->name != NULL; p++) { b ; }

#define MHASH_ALG_LOOP(a) \
	MHASH_LOOP( if(p->id == type) { a; break; } )

size_t
mhash_count(void)
{
	size_t count = 0;

	MHASH_LOOP(count++);

	return count;
}

size_t
mhash_get_block_size(hashid type)
{
	size_t ret = 0;

	MHASH_ALG_LOOP(ret = p->blocksize);
	return ret;
}

char *
mhash_get_hash_name(hashid type)
{
	char *ret = NULL;

	MHASH_ALG_LOOP(ret = strdup(p->name));
	return ret;
}

MHASH
mhash_init_int(const hashid type, const MHASH thread)
{
	if (!used_all_threads) {
		crc = realloc(crc, (thread + 1) * sizeof(word32));
		md5 = realloc(md5, (thread + 1) * sizeof(MD5_CTX));
		sha1 = realloc(sha1, (thread + 1) * sizeof(SHA1_CTX));
		gost = realloc(gost, (thread + 1) * sizeof(GostHashCtx));
		haval = realloc(haval, (thread + 1) * sizeof(haval_state));
		rmd128 = realloc(rmd128, (thread + 1) * sizeof(word32) * 4);
		rmd160 = realloc(rmd160, (thread + 1) * sizeof(word32) * 5);
		tiger_data = realloc(tiger_data, (thread + 1) * sizeof(word64) * 3);
		algorithm_given = realloc(algorithm_given, (thread + 1) * sizeof(hashid));
	}

	algorithm_given[thread] = type;
	check_pointer[thread] = 1;

	switch (type) {
	case CRC32:
	case CRC32B:
		clear_crc32(&crc[thread]);
		return thread;
	case MD5:
		MD5Init(&md5[thread]);
		return thread;
	case SHA1:
		SHA1Init(&sha1[thread]);
		return thread;
	case HAVAL:
		haval_start(&haval[thread]);
		return thread;
	case RIPEMD128:
		MDinit_128(&rmd128[thread]);
		return thread;
	case RIPEMD160:
		MDinit_160(&rmd160[thread]);
		return thread;
	case TIGER:
		return thread;
	case GOST:
		gosthash_reset(&gost[thread]);
		return thread;
	}
	return (MHASH) - 1;
}

/*
   plaintext should be in block's size 
 */
int
mhash(MHASH thread, const void *plaintext, size_t size)
{
	switch (algorithm_given[thread]) {
	case CRC32:
		crc32(&crc[thread], plaintext, size);
		break;
	case CRC32B:
		crc32b(&crc[thread], plaintext, size);
		break;
	case MD5:
		MD5Update(&md5[thread], plaintext, size);
		break;
	case SHA1:
		SHA1Update(&sha1[thread], plaintext, size);
		break;
	case HAVAL:
		haval_hash(&haval[thread], plaintext, size);
		break;
	case RIPEMD128:
		MDfinish_128(&rmd128[thread], plaintext, size, 0);
		break;
	case RIPEMD160:
		MDfinish_160(&rmd160[thread], plaintext, size, 0);
		break;
	case TIGER:
		tiger(plaintext, size, &tiger_data[thread * 3]);
		break;
	case GOST:
		gosthash_update(&gost[thread], plaintext, size);
		break;
	}
	return 0;
}


void *
mhash_end(MHASH thread)
{
	void *digest;

	check_pointer[thread] = 0;

	switch (algorithm_given[thread]) {
	case CRC32:
	case CRC32B:
		return get_crc32(&crc[thread]);
	case MD5:
		return MD5Final(&md5[thread]);
	case SHA1:
		return SHA1Final(&sha1[thread]);
	case HAVAL:
		digest = malloc(mhash_get_block_size(algorithm_given[thread]));
		haval_end(&haval[thread], digest);
		return digest;
	case RIPEMD160:
		digest = malloc(160 >> 3);
		memcpy(digest, &rmd160[thread], 160 >> 3);
		return digest;
	case RIPEMD128:
		digest = malloc(128 >> 3);
		memcpy(digest, &rmd128[thread], 128 >> 3);
		return digest;
	case TIGER:
		digest = malloc(192 >> 3);
		memcpy(digest, &tiger_data[thread * 3], 192 >> 3);
		return digest;
	case GOST:
		digest = malloc(32);
		gosthash_final(&gost[thread], digest);
		return digest;
	}

	return NULL;
}

MHASH
mhash_init(const hashid type)
{
	static int gost_init, crc32b_init;
	MHASH ret = MHASH_FAILED;

	CURRENT_COUNTER_LOCK;

	if (type == GOST && !gost_init) {
		gosthash_init();
		gost_init++;
	}

	if (type == CRC32B && !crc32b_init) {
		crc32b_init++;
		crc32bgen();
	}

	if (cur_thread == MAX_THREADS) {
		/*
		   FIXME: fixed size solutions can be improved 
		 */
		cur_thread = 0;
		used_all_threads = 1;
	}

	while (check_pointer[cur_thread] == 1 && cur_thread < MAX_THREADS) {
		cur_thread++;
	}

	if (cur_thread < MAX_THREADS) {
		ret = mhash_init_int(type, cur_thread);

		if (ret < 0) {
			mhash_end(cur_thread);
			ret = MHASH_FAILED;
		}
	}

	CURRENT_COUNTER_UNLOCK;

	return ret;
}
